const TimerStates = Object.freeze({
  PAUSED: 'paused',
  RUNNING: 'running',
  FINISHED: 'finished',
});

class TimerStateError extends Error {
  constructor(message) {
    super(message);
    this.name = 'TimerStateError';
  }
}

class CountdownTimer {
  constructor({ startTime, onFinish = () => {} }) {
    this.startTime = startTime;
    this.currentTime = startTime;
    this.onFinish = onFinish;
    this._timeoutIds = [];
    this._state = TimerStates.PAUSED;
  }

  start() {
    if (this._state === TimerStates.PAUSED) {
      this._state = TimerStates.RUNNING;
      this._timeoutIds.push(setTimeout(() => this._tick(), this._getTickPeriod()));
    } else {
      throw new TimerStateError(`Cannot start timer from state: '${this._state}'`);
    }
  }

  pause() {
    if (this._state === TimerStates.RUNNING) {
      this._state = TimerStates.PAUSED;
      this._clearTimeouts();
    } else {
      throw new TimerStateError(`Cannot pause timer from state: '${this._state}'`);
    }
  }

  finish() {
    if (this._state === TimerStates.RUNNING || this._state === TimerStates.PAUSED) {
      this._state = TimerStates.FINISHED;
      this.currentTime = 0;
      this._clearTimeouts();

      if (this.onFinish) {
        this.onFinish();
      }
    } else {
      throw new TimerStateError(`Cannot finish timer from state: '${this._state}'`);
    }
  }

  _clearTimeouts() {
    this._timeoutIds.forEach(clearTimeout);
  }

  _displayTime() {
    const maxDigitLength = this._getMaxDigitLength();
    const formattedTime = this._getFormattedTime();

    // Blank the current line and then display the current (formatted) time.
    process.stdout.write('\r' + ' '.repeat(maxDigitLength));
    process.stdout.write('\r' + formattedTime);
  }

  _getFormattedTime() {
    const numDecimals = this._getNumDecimals();

    const hours   = Math.floor( this.currentTime / 3600 ) % 24;
    const minutes = Math.floor( this.currentTime / 60 ) % 60;
    const seconds = (this.currentTime % 60).toFixed(numDecimals);

    let hoursStr = `${hours}`;
    let minutesStr = `${minutes}`;
    let secondsStr = `${seconds}`;

    if (hours < 10) {
      hoursStr = `0${hours}`;
    }

    if (minutes < 10 && hours >= 1) {
      minutesStr = `0${minutes}`;
    }

    if (seconds < 10 && (minutes >= 1 || hours >= 1)) {
      secondsStr = `0${seconds}`;
    }

    let timeParts = [];

    if (hours > 0) {
      timeParts.push(hoursStr);
    }

    if (minutes > 0 || hours > 0) {
      timeParts.push(minutesStr);
    }

    timeParts.push(secondsStr);

    return timeParts.join(':');
  }


  /*
   * Returns the number of miliseconds between ticks as a function of the
   * current time remaining. The closer the timer is to 0 the more frequently
   * we want to update the display. However, we never want the time between
   * redraws to exceed 1 second (1000 ms) or fall below 50 ms (to prevent flicker).
   */
  _getTickPeriod() {
    const tickScale = 1.1;

    return Math.max(Math.min(this.currentTime * tickScale, 1000), 50);
  }

  /*
   * Returns the number of extra decimal places to display when using sub-
   * second resolution.
   */
  _getExpandedDecimalLength() {
    return 1;
  }

  /*
   * Returns the number of extra decimal places to display for the current time.
   */
  _getNumDecimals() {
    return this.currentTime < 10 ? this._getExpandedDecimalLength() : 0;
  }

  /*
   * Returns the maximum string length that can be displayed by our timer.
   * This is necessary so that we can overwrite the time for display even as
   * its length changes.
   */
  _getMaxDigitLength() {
    const startHours = Math.floor( this.startTime / 3600 ) % 24;
    return startHours.toString().length + 1 +   // Hours and a colon
      2 + 1 +                                   // Minutes and a colon
      2 + 1 + this._getExpandedDecimalLength(); // Seconds, a decimal point, and decimals
  }

  _tick() {
    const start = Date.now();

    this._timeoutIds.push(setTimeout(() => {
      const duration = (Date.now() - start) / 1000;
      this.currentTime -= duration;

      if (this.currentTime <= 0) {
        console.log('');
        this.finish();
      } else {
        this._displayTime();
        this._tick();
      }
    }, this._getTickPeriod()));
  }
}

module.exports = CountdownTimer;
